"""
Module that facilitates several ways to send email notifications with gmail.

It also supports ignoring mail notifications, info should be a dictionary
with "user", "password", "recipients"
"""
import smtplib
import logging
import os
import traceback
from threading import Thread
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


def send(subject, body, info):
    """
    Basic synchronous email sending function given subject, body and information.
    """
    try:
        message = _createmessage(subject, body, info)
        server_ssl = _login(info)
        server_ssl.sendmail(info["user"],
                            _recipients(info),
                            message.as_string())
        logging.info('Notification "%s"', subject)
    except smtplib.SMTPException:
        logging.exception('Failed to send email %s', subject)
    except:
        logging.exception('Failed to send message %s', subject)


def asyncsend(subject, body, info):
    """
    Same as 'send', but performed in a background thread.

    The thread is automatically started, but not joined on.
    Still, the thread object is returned in case it is
    required.
    """
    thread = Thread(target=send, args=(subject, body, info))
    thread.start()
    return thread


def sendexception(exception, info, tracebacktxt=None, async=True, extra_subject=''):
    """
    Deliver an email with useful exception information.

    If no formatted traceback is specified, the current
    one is used. The async parameter is for calling asyncsend
    or send, respectively.

    You can add details to the subject with the variable extra_subject
    """
    subject = 'An exception occurred ' + extra_subject
    body = """
    An exception has ocurred:
    
    %s

    %s
    """ % (exception.message,
           tracebacktxt or traceback.format_exc())
    if async:
        asyncsend(subject, body, info)
    else:
        send(subject, body, info)


def send500(exception, tracebacktxt, request, info, extra_subject=''):
    """
    Asynchronously deliver a message in case of a 500 error.

    It uses information from the request, the
    exception object and the non-optional fromatted
    traceback.
    You can add details to the subject with the variable extra_subject
    """
    subject = '500 error ' + extra_subject
    body = """
    Internal server error at %s
    
    Arguments:
    %s

    Exception '%s':

    %s
    """ % (request.path, repr(dict(request.args)),
           exception.message, tracebacktxt)

    asyncsend(subject, body, info)


def _loadcredentials(info):
    """ 
    Load the credentials of the mail account from a dictionary
    """
    return info["user"], info["password"]


def _recipients(info):
    return info["recipients"].split(',')


def _recipientsstr(info):
    return ', '.join(_recipients(info))


def _createmessage(subject, body, info):
    """
    Create a message with the valid sintax
    """
    message = MIMEMultipart()
    message['Subject'] = subject
    message['From'] = info["user"]
    message['To'] = _recipientsstr(info)
    message.attach(MIMEText(body, 'plain', 'utf-8'))
    return message


def _login(info):
    """
    Logs to the mail account given by the user
    """
    user, password = _loadcredentials(info)
    server_ssl = smtplib.SMTP_SSL('smtp.gmail.com', 465)
    server_ssl.ehlo()
    server_ssl.login(user, password)
    return server_ssl



"""
Module for containing all required initialization logic.

It exposes a function that should ALWAYS be indepotent,
which means that it can be called one or possible multiple
times without creating fatal errors (other than those originated
by latency.)
"""
import sys
import logging
import logging.config
import os
from os.path import join, isfile
from dotenv import load_dotenv


def loads(**kwargs):
    """Load and check all the necessary configuration options"""

    # Load environment variables from envfile if present
    envPath = kwargs.pop('envPath', os.getcwd())
    env = kwargs.pop('envFile','.env')
    ENVFILE_PATH = join(envPath, env)
    if isfile(ENVFILE_PATH):
        load_dotenv(ENVFILE_PATH)
    else:
        logging.error('No .env file found')

    # Load logging setup
    logPath = kwargs.pop('logPath', os.getcwd())
    LOGFILE_PATH = join(logPath, 'logging.conf')
    if isfile(LOGFILE_PATH):
        logging.config.fileConfig(LOGFILE_PATH)
    else:
        logging.error('No logging.conf file found')

    productionOnly = set(kwargs.pop('productionOnly', {}))

    alwaysRequired = set(kwargs.pop('alwaysRequired', {}))

    #Validate no extra arguments
    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    reqs = set()

    if os.getenv('SERVICE_ENV') == 'dev':
        required = reqs.union(alwaysRequired)
    else:
        required = reqs.union(productionOnly | alwaysRequired)

    real = set(os.environ.keys())

    bad = False
    diff = required - real
    if len(diff) > 0:
        bad = True
        logging.error('Missing environment variables: %s', ', '.join(diff))

    if bad:
        sys.exit(1)
    else:
        logging.info('All environment variables are valid.')



"""
Module for testing validation.py

This module si for testing each use case for the validations
Since the main class Validation to the sub classes and the
validate function with this two cases.
"""
import pytest
from attendant.validation import Validation, Multivalue, MultivalueAll, Univalue, UnivaluePosType, validate

def test_call_classMultivalue_pass():
	a = Multivalue(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	assert callable(a)


def test_Multivalue_pass():
	a = Multivalue(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	assert a("0, 1", int) == ''

def test_Multivalue_fail():
	a = Multivalue(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	assert a("0,4,6", int) == '0 or/and 1 was invalid'


def test_Multivalue_fail2():
	a = Multivalue(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	assert a("0,1", str) == '0 or/and 1 was invalid'


def test_call_classMultivalueAll_pass():
	a = MultivalueAll(lambda xs: all(x in (0, 1, "all") for x in xs), "0 or/and 1 or/and all")
	assert callable(a)


def test_MultivalueAll_pass():
	a = MultivalueAll(lambda xs: all(x in (0, 1, "all") for x in xs), "0 or/and 1 or/and all")
	assert a("0,all", int) == ''


def test_MultivalueAll_pass1():
	a = MultivalueAll(lambda xs: all(x in (0, 1, "all") for x in xs), "0 or/and 1 or/and all")
	assert a("all", int) == ''


def test_MultivalueAll_fail():
	a = MultivalueAll(lambda xs: all(x in (0, 1, "all") for x in xs), "0 or/and 1 or/and all")
	assert a("0,4,and", int) == "0 or/and 1 or/and all was invalid"


def test_MultivalueAll_fail2():
	a = MultivalueAll(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	assert a("0,1,all", str) == '0 or/and 1 was invalid'


def test_call_classUnivalue_pass():
	a = Univalue(lambda x: x in (0,1), "0 or 1")
	b = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	assert callable(a)
	assert callable(b)


def test_Univalue_pass():
	a = Univalue(lambda x: x in (0,1), "0 or 1")
	assert a(0, int) == ''


def test_Univalue_pass2():
	a = Univalue(lambda x: x in (0,1), "0 or 1")
	assert a(1, int) == ''
	

def test_Univalue_fail():
	a = Univalue(lambda x: x in (0,1), "0 or 1")
	assert a("hello", str) == "0 or 1 was invalid"


def test_Univalue_fail2():
	a = Univalue(lambda x: x in (0,1), "0 or 1")
	assert a("0,1", int) == "0 or 1 was invalid"


def test_Univalue_fail3():
	a = Univalue(lambda x: x in (0,1), "0 or 1")
	assert a(3, int) == "0 or 1 was invalid"


def test_Univalue_str_pass():
	a = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	assert a("hola", str) == ''


def test_Univalue_str_pass2():
	a = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	assert a("adios", str) == ''


def test_Univalue_str_fail():
	a = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	assert a("hol", str) == "hola or adios was invalid"


def test_Univalue_str_fail1():
	a = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	assert a(2, str) == "hola or adios was invalid"


def test_Univalue_str_fail2():
	a = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	assert a(3, int) == "hola or adios was invalid"


def test_call_classUnivaluePosType_pass():
	a = UnivaluePosType("positive or type")
	assert callable(a)


def test_UnivaluePosType_pass():
	a = UnivaluePosType("positive or type")
	assert a(1, int) == ''


def test_UnivaluePosType_pass2():
	a = UnivaluePosType("positive or type")
	assert a("hello", str) == ''


def test_UnivaluePosType_fail():
	a = UnivaluePosType("positive or type")
	assert a(-1, int) == 'positive or type was invalid'


def test_UnivaluePosType_fail2():
	a = UnivaluePosType("positive or type")
	assert a("bye", int) == 'positive or type was invalid'


def test_validate_Multiclass_pass():
	a = Multivalue(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	b = MultivalueAll(lambda xs: all(x in (0, 1, "all") for x in xs), "0 or/and 1 or/and all")
	c = Univalue(lambda x: x in (0,1), "0 or 1")
	d = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	e = UnivaluePosType("positive or type")

	params = {
		"0 or/and 1": ["0", int],
		"0 or/and 1 or/and all": ["1,all", int],
		"0 or 1": [1, int],
		"hola or adios": ["hola", str],
		"positive or type": [6, int]
	}

	schema = {
		"0 or/and 1": a,
		"0 or/and 1 or/and all": b,
		"0 or 1": c,
		"hola or adios": d,
		"positive or type": e
	}

	errors = validate(schema, params)
	assert len(errors) == 0


def test_validate_Multiclass_fail():
	a = Multivalue(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	b = MultivalueAll(lambda xs: all(x in (0, 1, "all") for x in xs), "0 or/and 1 or/and all")
	c = Univalue(lambda x: x in (0,1), "0 or 1")
	d = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	e = UnivaluePosType("positive or type")

	params = {
		"0 or/and 1": ["0", int],
		"0 or/and 1 or/and all": ["1,all", int],
		"0 or 1": [1, int],
		"hola or adios": ["hola", int],
		"positive or type": [6, int]
	}

	schema = {
		"0 or/and 1": a,
		"0 or/and 1 or/and all": b,
		"0 or 1": c,
		"hola or adios": d,
		"positive or type": e
	}

	errors = validate(schema, params)
	assert len(errors) == 1
	assert errors == {'hola or adios': 'hola or adios was invalid'}


def test_validate_Multiclass_fail2():
	a = Multivalue(lambda xs: all(x in (0, 1) for x in xs), "0 or/and 1")
	b = MultivalueAll(lambda xs: all(x in (0, 1, "all") for x in xs), "0 or/and 1 or/and all")
	c = Univalue(lambda x: x in (0,1), "0 or 1")
	d = Univalue(lambda x: x in ("hola","adios"), "hola or adios")
	e = UnivaluePosType("positive or type")

	params = {
		"0 or/and 1": ["0", int],
		"0 or/and 1 or/and all": ["1,all", int],
		"0 or 1": [1, int],
		"hola or adios": ["hola", str],
		"positive or type": [-6, int]
	}

	schema = {
		"0 or/and 1": a,
		"0 or/and 1 or/and all": b,
		"0 or 1": c,
		"hola or adios": d,
		"positive or type": e
	}

	errors = validate(schema, params)
	assert len(errors) == 1
	assert errors == {'positive or type': 'positive or type was invalid'}


"""
Module that make validations depnding on parameters.

This module uses several classes to make different types of
validations.
"""
import logging

class Validation:
    """
    Father class for all the validations.

    Class that contain the constructor of the object and have
    the father callable method.

    Subclasses need to override the method _validation.
    """
    def __init__(self, lamb, name):
        if callable(lamb):
            self.lamb = lamb
        else:
            raise TypeError("lamb should be a lambda")
        if type(name) == str:
            self.name = name
        else:
            raise TypeError("name should be a string")


    def __call__(self, received, typefn):
        return self._validationstructure(self.name, received, typefn, self.lamb)


    def _valudationalgorithm(self, name, received, typefn, validatecasted):
            pass


    def _validationstructure(self, name, received, typefn, validatecasted):
        try:
            return self._valudationalgorithm(name, received, typefn, validatecasted)
        except (ValueError, TypeError) as e:
            logging.warning(self._errorMessage(name, str(e)))
            return self._errorMessage(name)


    def _errorMessage(self, name, exception=''):
        return (('%s: %s was invalid', str(exception), name) if exception 
                    else ('%s was invalid' % name))



class Multivalue(Validation):
    """
    Class for the Multivalue validation.

    This class implements the Validation class, it
    override the _validation method with the local
    method _multivalued.
    """
    def _valudationalgorithm(self, name, received, typefn, validatecasted):
        values = received.split(',')
        casted = [typefn(value) for value in values]
        return '' if validatecasted(casted) else (self._errorMessage(name))


class MultivalueAll(Multivalue):
    """
    Class for the Multivalue_all validation.

    This class implements the Multivalue class
    and overrides the _validation method with thefather 
    method _multivalued using the local method _typeorall.
    """
    def __call__(self, received, typefn):
        return self._validationstructure(self.name, received, self._typeorall(typefn), self.lamb)


    def _typeorall(self, fn):
        def callback(value):
            if value == 'all':
                return value
            else:
                return fn(value)
        return callback


class Univalue(Validation):
    """
    Class for the Univalue validation.

    This class implements the Validation class, it
    override the _validation method with the local
    method _valued.
    """
    def _valudationalgorithm(self, name, received, typefn, validatecasted):
        if type(received) == typefn:
            return '' if validatecasted(received) else (self._errorMessage(name))
        else:
            raise TypeError


class UnivaluePosType(Validation):
    """
    Class for the Univalue_pos_type validation.

    This class implements the Validation class, it
    override the _validation method with the local
    method _typed.
    """
    def __init__(self, name):
        self.lamb = lambda x:True
        if type(name) == str:
            self.name = name
        else:
            raise TypeError("name should be a string")

    def _valudationalgorithm(self, name, received, typefn, lamb):
        typefn(received)
        if typefn is int and int(received) < 0:
            raise ValueError(name + ' cannot be negative')
        return ''


def validate_keys(schema, params):
    return list(set(schema.keys()) - set(params.keys()))


def validate(schema, params):
    """
    Function that validates a schema with parameters
    
    Validates each entry of params against the cooresponding entry of the
    schema using functions in this file.
    """
    errors = {}
    missing = validate_keys(schema, params)

    if len(missing) > 0:
        mistrs = ''
        for k in missing[:-1]:
            mistrs += str(k) + ', '
        mistrs += str(missing[len(missing) - 1])
        logging.warning('There is/are %d missing keys: %s.',
                        len(missing), mistrs)
        errors['missing'] = missing

    for key, validator in schema.iteritems():
        if key not in params:
            continue
        error = validator(params[key][0], params[key][1])
        errors.update({key: error} if error else {})

    logging.warning('There were %d errors.', len(errors))
    return errors

